/*
    Sequencer.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <cstdio>

#include <iostream>

#include "Sequencer.h"
#include "instrument.h"
#include "MIDI.h"

using namespace std;

const int Sequencer::block2Part[32] =
{
  9, 0, 1, 2, 3, 4, 5, 6, 7, 8,10,11,12,13,14,15,
  25,16,17,18,19,20,21,22,23,24,26,27,28,29,30,31
};

const int Sequencer::block2Port[32] =
{
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,//MIDI::PORT_A
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,//MIDI::PORT_B
};

const int Sequencer::block2Channel[32] =
{
  9, 0, 1, 2, 3, 4, 5, 6, 7, 8,10,11,12,13,14,15,
  9, 0, 1, 2, 3, 4, 5, 6, 7, 8,10,11,12,13,14,15
};


Sequencer::Sequencer (State* globalState, Option* midiOption):
  option (midiOption), state (globalState), cacheState (*globalState),
  isSeeking_ (false)
{
  reset ();
}

Sequencer::~Sequencer ()
{
  //do nothing
}

void Sequencer::resetPartData ()
{
  for (int port = 0; port < MIDI::MAX_PORT; port++)
    {
      for (int channel = 0; channel < MIDI::MAX_CHANNEL; channel++)
	{
	  cacheState.part[port][channel].reset(port, channel);
	}
    }
}

void Sequencer::changeRPN_NRPN (bool isRPN,
				int port, uchar channel, uchar msb, uchar lsb,
				uchar cc1, uchar cc1Value,
				uchar cc2, uchar cc2Value)
{
  assert (0<= port);
  assert (port < MIDI::MAX_PORT);
  assert (channel < MIDI::MAX_CHANNEL);
  assert ((cc1 == 0x06 && cc2 == 0x26) ||
	  (cc1 == 0x60 && cc1Value == 0) ||
	  (cc1 == 0x61 && cc1Value == 0));

  const int savedParameterKind = cacheState.part[port][channel].getParameterKind ();
  const uchar savedRpnMSB = cacheState.part[port][channel].getRpnMSB ();
  const uchar savedRpnLSB = cacheState.part[port][channel].getRpnLSB ();
  const uchar savedNrpnMSB = cacheState.part[port][channel].getNrpnMSB ();
  const uchar savedNrpnLSB = cacheState.part[port][channel].getNrpnLSB ();

  if (isRPN)
    {
      //RPN$B$N;XDj$NHV9f$K$9$k!#(B
      if (savedRpnMSB != msb)
	controlChange (port, channel, 0x65, msb);
      if ((savedParameterKind != Part::PARAMETER_RPN && savedRpnMSB == msb) ||
	  savedRpnLSB != lsb)
	controlChange (port, channel, 0x64, lsb);
    }
  else
    {
      //NRPN$B$N;XDj$NHV9f$K$9$k!#(B
      if (savedNrpnMSB != msb)
	controlChange (port, channel, 0x63, msb);
      if ((savedParameterKind != Part::PARAMETER_NRPN && savedNrpnMSB == msb)||
	  savedNrpnLSB != lsb)
	controlChange (port, channel, 0x62, lsb);
    }

  //RPN or NRPN$B$rA`:n(B
  if (cc1 == 0x06)
    {//data entry
      controlChange (port, channel, cc1, cc1Value);
      controlChange (port, channel, cc2, cc2Value);
    }
  else
    {//incriment or decriment
      controlChange (port, channel, cc1, cc1Value);
    }
  

  if (savedParameterKind == Part::PARAMETER_NULL)
    {
      controlChange (port, channel, 0x65, 0x7f);
      controlChange (port, channel, 0x64, 0x7f);
    }
  else if (isRPN)
    {
      //$B85$KLa$9!#(B
      if (savedParameterKind == Part::PARAMETER_RPN)
	{
	  if (savedRpnMSB != msb)
	    controlChange (port, channel, 0x65, savedRpnMSB);
	  if (savedRpnLSB != lsb)
	    controlChange (port, channel, 0x64, savedRpnLSB);
	}
      else if (savedParameterKind == Part::PARAMETER_NRPN)
	{//NRPN$B$K$7$F$*$1$P$$$$$N$G!"(BMSB$B$@$1Aw$k!#(B
	  controlChange (port, channel,
			 0x63, cacheState.part[port][channel].getNrpnMSB ());
	}
    }
  else
    {
      //$B85$KLa$9!#(B
      if (savedParameterKind == Part::PARAMETER_RPN)
	{//NRPN$B$K$7$F$*$1$P$$$$$N$G!"(BMSB$B$@$1Aw$k!#(B
	  controlChange (port, channel,
			 0x65, cacheState.part[port][channel].getRpnMSB ());
	}
      else if (savedParameterKind == Part::PARAMETER_NRPN)
	{
	  if (savedNrpnMSB != msb)
	    controlChange (port, channel, 0x65, savedNrpnMSB);
	  if (savedNrpnLSB != lsb)
	    controlChange (port, channel, 0x64, savedNrpnLSB);
	}
    }
}

//////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////trace MIDI status
//////////////////////////////////////////////////////////////////////////
//port==0 port_A
//port==1 port_B
void Sequencer::tracePortChange (int port)
{
  //DEBUG_MESSAGE ("Sequencer::portChange port change meta Event found.ingnoring.sorry not supported yet.");
  switch (option->portChangeType)
    {
    case Option::IGNORE:
      //do nothing
      break;
    case Option::SERIAL:
      if (port == 0)
	cacheState.sequencerStatus.setPort (MIDI::PORT_A);
      else if (port == 1)
	cacheState.sequencerStatus.setPort (MIDI::PORT_B);
      break;
    case Option::USE_ROLAND_SYS_EX:
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$BL$<BAu(B
      break;
    }
}


void Sequencer::traceNoteOff(int port, uchar channel,
			     uchar note, uchar velocity)
{
  assert (note < 128);
  assert (velocity < 128);
  
  for (int port__ = 0; port__ < MIDI::MAX_PORT; port__++)
    {
      for (int channel__ = 0; channel__ < MIDI::MAX_CHANNEL; channel__++)
	{
	  if (cacheState.part[port__][channel__].isRecieve (port, channel))
	    cacheState.part[port__][channel__].traceNoteOff (note);
	}
    } 
}


void Sequencer::traceNoteOn (int port, uchar channel,
			     uchar note, uchar velocity)
{
  assert (note < 128);
  assert (velocity < 128);
  
  for (int port__ = 0; port__ < MIDI::MAX_PORT; port__++)
    {
      for (int channel__ = 0; channel__ < MIDI::MAX_CHANNEL; channel__++)
	{
	  if (cacheState.part[port__][channel__].isRecieve (port, channel))
	    cacheState.part[port__][channel__].traceNoteOn (note, velocity,
							    cacheState.songStatus.getMachineTicks ());
	}
    } 
}

void Sequencer::tracePolyphonicKeyPressure (int port, uchar channel,
					    uchar note, uchar keyPressure)
{
  assert (note < 128);
  assert (keyPressure < 128);
  
  for (int port__ = 0; port__ < MIDI::MAX_PORT; port__++)
    {
      for (int channel__ = 0; channel__ < MIDI::MAX_CHANNEL; channel__++)
	{
	  if (cacheState.part[port__][channel__].isRecieve (port, channel))
	    cacheState.part[port__][channel__].tracePolyphonicKeyPressure(note, keyPressure);
	}
    } 
}

void Sequencer::traceControlChange (int port, uchar channel,
				    uchar control, uchar value)
{
  assert (control < 128);
  assert (value < 128);

  for (int port__ = 0; port__ < MIDI::MAX_PORT; port__++)
    {
      for (int channel__ = 0; channel__ < MIDI::MAX_CHANNEL; channel__++)
	{
	  if (cacheState.part[port__][channel__].isRecieve (port, channel))
	    cacheState.part[port__][channel__].traceControlChange (control,
								   value);
	}
    } 
}


void Sequencer::traceProgramChange (int port, uchar channel, uchar program)
{
  assert (program<128);
  
  for (int port__ = 0; port__ < MIDI::MAX_PORT; port__++)
    {
      for (int channel__ = 0; channel__ < MIDI::MAX_CHANNEL; channel__++)
	if (cacheState.part[port__][channel__].isRecieve (port, channel))
	  cacheState.part[port__][channel__].traceProgramChange (program, cacheState.sequencerStatus.getDeviceMode ());
    } 
}

void Sequencer::traceChannelPressure (int port, uchar channel, uchar pressure)
{
  assert (pressure < 128);
  
  for (int port__ = 0; port__ < MIDI::MAX_PORT; port__++)
    {
      for (int channel__ = 0; channel__ < MIDI::MAX_CHANNEL; channel__++)
	if (cacheState.part[port__][channel__].isRecieve (port, channel))
	  cacheState.part[port__][channel__].traceChannelPressure (pressure);
    } 
}


void Sequencer::tracePitchBendChange (int port, uchar channel,
				      uchar lsb, uchar msb)
{
  assert (lsb < 128);
  assert (msb < 128);
  
  for (int port__ = 0; port__ < MIDI::MAX_PORT; port__++)
    {
      for (int channel__ = 0; channel__ < MIDI::MAX_CHANNEL; channel__++)
	if (cacheState.part[port__][channel__].isRecieve (port, channel))
	  cacheState.part[port__][channel__].tracePitchBendChange (lsb, msb);
    } 
}

void Sequencer::traceExclusive (const uchar* udata, uint32 size,
				int port = -1, int channel = -1)
{
  assert (size >= 5);
  //exclusive:0xF0 ID deviceID data ... 0xF7
  //#ifdef DEBUG
#if 0
#if 1
  cerr << "exclusive message=" << endl;
  dumpgEvent (udata, size);
#else
  io.errorMessage ("eclusive message.");
  ostrstream message;
  dumpEvent (udata, size);
  errorMessage (message.str ());
  message.freeze (0);
#endif
#endif
  switch (udata[1])
    {
    case MIDI::UNIVERSAL_REALTIME:
      traceUniversalRealtimeExclusive (udata, size);
      break;
    case MIDI::UNIVERSAL_NONREALTIME:
      traceUniversalNonrealtimeExclusive (udata, size);
      break;
    case MIDI::ROLAND_ID:
      traceRolandExclusive (udata, size);
      break;
    case MIDI::YAMAHA_ID:
      traceYamahaExclusive (udata, size);
      break;
    default:
      //do nothing
#ifdef DEBUG
      cerr << "unkown exclusive ID=" << hex << static_cast<int>(udata[1])
	   << dec <<endl;
      dumpEvent (udata, size);
#endif
      break;
    }
}

void Sequencer::traceUniversalRealtimeExclusive (const uchar* udata,
						 uint32 size)
{
  if (size < 5)
    return;//too short universal exclusive message

  //support$B0J30$OL5;k!#(B

  //master volume
  if (size == 8        && //??? need this?
      udata[2] == 0x7F && //broad cast
      udata[3] == 0x04 && //subid1 device control message
      udata[4] == 0x01 && //subid2 master volume
      udata[7] == 0xF7    //end of exclusive //??? need this? or check size?
      )
    {
      cacheState.sequencerStatus.setMasterVolume (udata[5]+(udata[6] << 7));
      return;
    }
  
#ifdef DEBUG
  cerr << "unkown universal real time message id="
       << hex<< static_cast<int>(udata[2]) << ":" << (int)udata[3]
       << dec <<endl;
  dumpEvent (udata, size);
#endif
  return;
}

void Sequencer::traceUniversalNonrealtimeExclusive (const uchar* udata,
						    uint32 size)
{
  //GM SYSTEM ON/OFF
  if (size == 6        && //need this???
      udata[2] == 0x7F && //broad cast
      udata[3] == 0x09 && //sub id#1(General MIDI Message)
      udata[5] == 0xF7    //end of exclusive //??? need this?
      )
    {
      //sub id#2(General MIDI Message,0x01:ON,0x02:OFF)
      if (udata[4] == 0x01)//GM ON
	{
	  cacheState.sequencerStatus.setDeviceMode (cacheState.sequencerStatus.getDeviceMode () | SequencerStatus::DEVICE_GM_ON);
	  return;
	}
      if (udata[4] == 0x02)//GM OFF
	{
	  cacheState.sequencerStatus.setDeviceMode (cacheState.sequencerStatus.getDeviceMode () & ~SequencerStatus::DEVICE_GM_ON);
	  return;
	}
    }

#ifdef DEBUG
  cerr << "unkown universal non-real time message id="
       << hex<< static_cast<int> (udata[2]) << ":"
       << static_cast<int> (udata[3]) << dec <<endl;
  dumpEvent (udata, size);
#endif
}

//0xF0 0x41(ROLAND) 0x10(device id) 
void Sequencer::traceRolandExclusive (const uchar* udata, uint32 size)
{
  assert (size >= 8);
  if (udata[4] != MIDI::DT1)
    {
#ifdef DEBUG
      cerr << "DEBUG:SysEx:Roland:unkown command id." << endl;
#endif
      return;
    }
  
  switch (udata[3])//model ID
    {
    case 0x42://system parameter
      traceRolandSystemParameter (udata, size);
      break;
    case 0x45://display data
      traceRolandDisplayData (udata, size);
      break;
    default:
#ifdef DEBUG
      cerr << "DEBUG:SysEx:Roland:unkown model id="
	   << hex << static_cast<int>(udata[3]) << dec << endl;
#endif
      break;
    }
}

//0xF0 0x41(ROLAND) 0x10(device id) 0x45(display data) 0x12(command id DT1)
//0x10 (address high)
void Sequencer::traceRolandDisplayData (const uchar* udata, uint32 size)
{
  assert (size >= 11);
  assert (udata[5] == 0x10);//address$B>e0L%P%$%H(B
  switch (udata[6])//address$BCf0L%P%$%H(B
    {
    case 0x00://$BJ8;zNs(B
      //Displayed Letter
      traceRolandDisplayedLetter (udata, size);
      break;
    case 0x01://dot pattern(Displayed Dot Data)
    case 0x02:
    case 0x03:
    case 0x04:
    case 0x05:
      traceRolandDisplayedDotData (udata, size);
      break;
    case 0x20://$BI=<(%Z!<%8(BDispay Page
      traceRolandDisplayPage (udata, size);
      break;
    default:
#ifdef DEBUG
      cerr << "DEBUG:roland::rolandDisplayData:unkown address!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
      dumpEvent (udata, size);
#endif
      break;
    }
}

//0xF0 0x41(ROLAND) 0x10(device id) 0x45(display data) 0x12(command id DT1)
//0x10(address high) 0x00 (address middle) 0x00 (address low)
void Sequencer::traceRolandDisplayedLetter (const uchar* udata, uint32 size)
{
  //assert (udata[7]==0x00);
  //assert (size==5+3+32+2);//command=5 address=3 data=32 checksum+EOX=2
  //assert (udata[7]==0x00);//address$B2<0L%P%$%H(B
  int i;
  for (i = 0; udata[8+i+1] != MIDI::EOX; i++)//[]$BFb$N(B+1$B$O(Bcheck sum$B$NJ,!#(B
    {
      assert(32 <= udata[8+i]||udata[8+i] <= 127);
      //cacheState.sequencerStatus.displayedLetter[i]=udata[8+i];
      cacheState.sequencerStatus.setDisplayedLetter (i, udata[8+i]);
    }
  cacheState.sequencerStatus.setDisplayedLetter (i, '\0');
#if 0
  cacheState.sequencerStatus.letterTicks = cacheState.songStatus.getTicks ();
  cacheState.sequencerStatus.letterMachineTicks = cacheState.songStatus.getMachineTicks ();
#endif
  cacheState.sequencerStatus.setLetterTicks (cacheState.songStatus.getTicks (),
					     cacheState.songStatus.getMachineTicks ());
#ifdef DEBUG
  //cerr << "DEBUG:Sequencer:display letter.size=" << size-8 << endl;
  //cerr << "letter=" << cacheState.sequencerStatus.displayedLetter << endl;
#endif
}

//0xF0 0x41(ROLAND) 0x10(device id) 0x45(display data) 0x12(command id DT1)
//0x10(address high) 0x0p (address middle,p=0x01-0x05) 
void Sequencer::traceRolandDisplayedDotData (const uchar* udata, uint32 size)
{
  assert (size == 5+3+64+2);//command=5 address=3 data=64 checksum+EOX=2
  assert (udata[7] == 0x00|| udata[7] == 0x40);//address$B2<0L%P%$%H(B
  //$B>e0L(B $BCf0L(B $B2<0L(B
  //0x10 0x0p {0x00, 0x40}
  //p=1-5
  const int page = (udata[6]-1)*2+((udata[7] == 0x40)?0:1)-1;//0<=page<=9
  cacheState.sequencerStatus.setChangedPage (page);
  uchar sum = 0;
  for (int i = 0;i<64;i++)
    {
      //cacheState.sequencerStatus.dotData[page][i] = udata[8+i];
      cacheState.sequencerStatus.setDotData (page, i, udata[8+i]);
      sum += udata[8+i];
    }
  sum &= ~((uchar)0x0F);
  //assert (sum == udata[8+64+1]);
#if 0
  cacheState.sequencerStatus.dotTicks = cacheState.songStatus.getTicks ();
  cacheState.sequencerStatus.dotMachineTicks =
    cacheState.songStatus.getMachineTicks ();
#endif
  cacheState.sequencerStatus.setDotTicks (cacheState.songStatus.getTicks (),
					  cacheState.songStatus.getMachineTicks ());
}


//0xF0 0x41(ROLAND) 0x10(device id) 0x45(display data) 0x12(command id DT1)
//0x10(address high) 0x20 (address middle) 0x00 (address low) 
void Sequencer::traceRolandDisplayPage (const uchar* udata, uint32 size)
{
  assert (size == 5+3+1+2);//command=5 address=3 data=1 checksum+EOX=2
  assert (udata[7] == 0x00 || udata[7] == 0x01);//address$B2<0L%P%$%H(B
  if (udata == 0x00)
    {
      assert (/*0x00 <= udata[8] &&*/ udata[8] <= 0x0A);
      cacheState.sequencerStatus.setDisplayPage (udata[8]);
    }
  else
    {
      assert (/*0x00 <= udata[8] &&*/ udata[8] <= 0x0F);
      cacheState.sequencerStatus.setDisplayTime (udata[8]);
    }
#if 0
  cacheState.sequencerStatus.pageTicks = cacheState.songStatus.getTicks ();
  cacheState.sequencerStatus.pageMachineTicks =
    cacheState.songStatus.getMachineTicks ();
#endif
  cacheState.sequencerStatus.setPageTicks (cacheState.songStatus.getTicks (),
					   cacheState.songStatus.getMachineTicks ());
}


void Sequencer::traceRolandSystemParameter (const uchar* udata, uint32 size)
{
#ifdef DEBUG
  /*
  cerr << "DEBUG:Sequencer:rolandSystemParameter" << endl;
  dumpEvent (udata,size);
  */
#endif
  const uint32 address = (udata[5] << 16)|(udata[6] << 8)|udata[7];
  //                    $B>e0L%"%I%l%9(B    $BCf0L%"%I%l%9(B    $B2<0L%"%I%l%9(B
  switch (address)
    {
    case 0x00007F://SYSTEM MODE SET
      assert (size == 5+3+1+2);//command=5 address=3 data=1 checksum+EOX=2
      //cacheState.sequencerStatus.mode=udata[8];
      cacheState.sequencerStatus.setDeviceMode (cacheState.sequencerStatus.getDeviceMode () | SequencerStatus::DEVICE_SC88_ON);
      break;
    case 0x000100://CHANNEL MSG RX PORT//$BD>2<$X(B
    case 0x000101://$BD>2<$X(B
    case 0x000102://$BD>2<$X(B
    case 0x000103://$BD>2<$X(B
    case 0x000104://$BD>2<$X(B
    case 0x000105://$BD>2<$X(B
    case 0x000106://$BD>2<$X(B
    case 0x000107://$BD>2<$X(B
    case 0x000108://$BD>2<$X(B
    case 0x000109://$BD>2<$X(B
    case 0x00010A://$BD>2<$X(B
    case 0x00010B://$BD>2<$X(B
    case 0x00010C://$BD>2<$X(B
    case 0x00010D://$BD>2<$X(B
    case 0x00010E://$BD>2<$X(B
    case 0x00010F://$BD>2<$X(B
    case 0x000110://$BD>2<$X(B
    case 0x000111://$BD>2<$X(B
    case 0x000112://$BD>2<$X(B
    case 0x000113://$BD>2<$X(B
    case 0x000114://$BD>2<$X(B
    case 0x000115://$BD>2<$X(B
    case 0x000116://$BD>2<$X(B
    case 0x000117://$BD>2<$X(B
    case 0x000118://$BD>2<$X(B
    case 0x000119://$BD>2<$X(B
    case 0x00011A://$BD>2<$X(B
    case 0x00011B://$BD>2<$X(B
    case 0x00011C://$BD>2<$X(B
    case 0x00011D://$BD>2<$X(B
    case 0x00011E://$BD>2<$X(B
    case 0x00011F:
      assert (size == 5+3+1+2);//command=5 address=3 data=1 checksum+EOX=2
      //XXX
      //cacheState.part[block2Part[udata[7]]].port = udata[8];
      cacheState.part[block2Port[udata[7]]][block2Channel[udata[7]]].setPort (udata[8]);
    break;
    case 0x400004://MASTER VOLUME
      assert (size == 5+3+1+2);//command=5 address=3 data=1 checksum+EOX=2
      cacheState.sequencerStatus.setMasterVolume ((udata[8] << 7)+0x7F);
      break;
    case 0x400005://MASTER KEY-SHIFT
      break;
    case 0x400006://MASTER PAN
      assert (size == 5+3+1+2);//command=5 address=3 data=1 cechsum+EOX=2
      cacheState.sequencerStatus.setMasterPan (udata[8]-0x40);
      break;
    case 0x40007F://MODE SET(GS RESET)
      cacheState.sequencerStatus.setDeviceMode (cacheState.sequencerStatus.getDeviceMode () | SequencerStatus::DEVICE_GS_ON);
      break;
    case 0x400100://PATCH NAME
      traceRolandDisplayedLetter (udata, size);
      break;
    case 0x401002://Rx.CHANNEL//$BD>2<$X(B
    case 0x401102://$BD>2<$X(B
    case 0x401202://$BD>2<$X(B
    case 0x401302://$BD>2<$X(B
    case 0x401402://$BD>2<$X(B
    case 0x401502://$BD>2<$X(B
    case 0x401602://$BD>2<$X(B
    case 0x401702://$BD>2<$X(B
    case 0x401802://$BD>2<$X(B
    case 0x401902://$BD>2<$X(B
    case 0x401A02://$BD>2<$X(B
    case 0x401B02://$BD>2<$X(B
    case 0x401C02://$BD>2<$X(B
    case 0x401D02://$BD>2<$X(B
    case 0x401E02://$BD>2<$X(B
    case 0x401F02://$BD>2<$X(B
    case 0x501002://Rx.CHANNEL//$BD>2<$X(B
    case 0x501102://$BD>2<$X(B
    case 0x501202://$BD>2<$X(B
    case 0x501302://$BD>2<$X(B
    case 0x501402://$BD>2<$X(B
    case 0x501502://$BD>2<$X(B
    case 0x501602://$BD>2<$X(B
    case 0x501702://$BD>2<$X(B
    case 0x501802://$BD>2<$X(B
    case 0x501902://$BD>2<$X(B
    case 0x501A02://$BD>2<$X(B
    case 0x501B02://$BD>2<$X(B
    case 0x501C02://$BD>2<$X(B
    case 0x501D02://$BD>2<$X(B
    case 0x501E02://$BD>2<$X(B
    case 0x501F02:      
      assert (size == 5+3+1+2);//command=5 address=3 data=2 checksum+EOX=2
      //cacheState.part[block2Part[udata[7]]].RxChannel = udata[8];
      {
	//XXX
	int port;
	if (cacheState.sequencerStatus.getPort () == MIDI::PORT_A)
	  port = (udata[5] == 0x40)? MIDI::PORT_A: MIDI::PORT_B;
	else
	  port = (udata[5] == 0x40)? MIDI::PORT_B: MIDI::PORT_A;//port == MIDI::PORT_B
	cacheState.part[port][block2Channel[(udata[6]&0x0F)]].setRxChannel (udata[8]);
      }
      break;
    case 0x401015://USE FOR RHYTHM PART//$BD>2<$X(B
    case 0x401115://$BD>2<$X(B
    case 0x401215://$BD>2<$X(B
    case 0x401315://$BD>2<$X(B
    case 0x401415://$BD>2<$X(B
    case 0x401515://$BD>2<$X(B
    case 0x401615://$BD>2<$X(B
    case 0x401715://$BD>2<$X(B
    case 0x401815://$BD>2<$X(B
    case 0x401915://$BD>2<$X(B
    case 0x401A15://$BD>2<$X(B
    case 0x401B15://$BD>2<$X(B
    case 0x401C15://$BD>2<$X(B
    case 0x401D15://$BD>2<$X(B
    case 0x401E15://$BD>2<$X(B
    case 0x401F15://$BD>2<$X(B
    case 0x501015://USE FOR RHYTHM PART//$BD>2<$X(B
    case 0x501115://$BD>2<$X(B
    case 0x501215://$BD>2<$X(B
    case 0x501315://$BD>2<$X(B
    case 0x501415://$BD>2<$X(B
    case 0x501515://$BD>2<$X(B
    case 0x501615://$BD>2<$X(B
    case 0x501715://$BD>2<$X(B
    case 0x501815://$BD>2<$X(B
    case 0x501915://$BD>2<$X(B
    case 0x501A15://$BD>2<$X(B
    case 0x501B15://$BD>2<$X(B
    case 0x501C15://$BD>2<$X(B
    case 0x501D15://$BD>2<$X(B
    case 0x501E15://$BD>2<$X(B
    case 0x501F15:
      assert (size == 5+3+1+2);//command=5 address=3 data=1 checksum+EOX=2
      //cacheState.part[MIDI::PORT_A][(udata[6]&0x0F)].useForRhythmPart = udata[8];
      {
	//XXX
	int port;
	if (cacheState.sequencerStatus.getPort () == MIDI::PORT_A)
	  port = (udata[5] == 0x40)? MIDI::PORT_A: MIDI::PORT_B;
	else
	  port = (udata[5] == 0x40)? MIDI::PORT_B: MIDI::PORT_A;//port == MIDI::PORT_B
	cacheState.part[port][block2Channel[(udata[6]&0x0F)]].setUseForRhythmPart (udata[8]);
      }
      break;
    case 0x400000://DRUM MAP NAME //$BD>2<$X(B
    case 0x401000:
    case 0x500000://DRUM MAP NAME //$BD>2<$X(B
    case 0x501000:
      traceRolandDisplayedLetter (udata, size);
      break;
    case 0x210000://USER DRUM SET NAME //$BD>2<$X(B
    case 0x211000:
      traceRolandDisplayedLetter (udata, size);
      break;
    case 0x320008://USER DRUM SET NAME //$BD>2<$X(B
    case 0x320108: //$BD>2<$X(B
    case 0x320208: //$BD>2<$X(B
    case 0x320308: //$BD>2<$X(B
    case 0x320408: //$BD>2<$X(B
    case 0x320508: //$BD>2<$X(B
    case 0x320608: //$BD>2<$X(B
    case 0x320708: //$BD>2<$X(B
    case 0x320808: //$BD>2<$X(B
    case 0x320908: //$BD>2<$X(B
    case 0x320A08: //$BD>2<$X(B
    case 0x320B08: //$BD>2<$X(B
    case 0x320C08: //$BD>2<$X(B
    case 0x320D08: //$BD>2<$X(B
    case 0x320E08: //$BD>2<$X(B
    case 0x320F08:
      traceRolandDisplayedLetter (udata, size);
      break;
    default:
      //do nothing
#ifdef DEBUG
      //cerr << "DEBUG:roland:system parameter:unkown address!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
#endif
      break;
    }
}

void Sequencer::traceYamahaExclusive (const uchar* udata, uint32 size)
{
  const uint32 address=((udata[4]) << 16) | (udata[5] << 8) | udata[6];
  switch (address)
    {
    case 0x00007E://XG System On
      cacheState.sequencerStatus.setDeviceMode (cacheState.sequencerStatus.getDeviceMode () | SequencerStatus::DEVICE_XG_ON);
      break;
    default:
#ifdef DEBUG
      cerr << "unkown universal YAMAHA time message id="
	   << hex << static_cast<int> (udata[2]) << ":"
	   << static_cast<int> (udata[3]) << dec << endl;
      dumpEvent (udata, size);
#endif
      break;
    }
}

void Sequencer::dumpEvent (const uchar* udata, uint32 size)
{
#ifdef DEBUG
  cerr << hex;
  for (uint i = 0; i < size; i++)
    cerr << static_cast<int>(udata[i]) << " ";
  cerr << dec << endl;
#endif
}
